@@ -1,4 +1,4 @@
-from typing import Dict, Literal, List, Optional
+from typing import Dict, List, Literal, Optional
 
 import pydantic
 
@@ -9,15 +9,12 @@ Conversation = List[ConversationDict]
 class ConversationTurn(pydantic.BaseModel):
     """
     Representation of a single turn in a conversation.
-
     This class defines a model for encapsulating a single conversational
     turn consisting of an input user's message and an output LLM message. It is
     designed to handle the exchange of messages in a structured format.
-
     Args:
         input: The input message of the conversation turn.
         output: The output message of the conversation turn.
-
     Example:
         >>> conversation_turn = ConversationTurn(
         >>>     input={"role": "user", "content": "Hello!"},
@@ -31,5 +28,7 @@ class ConversationTurn(pydantic.BaseModel):
     def as_list(self) -> List[ConversationDict]:
         if self.output is None:
             return [self.input]
-
         return [self.input, self.output]
+
+
+__all__ = ["ConversationDict", "Conversation", "ConversationTurn"]
