@@ -20,8 +20,8 @@ from . import (
 from .patchers import (
     litellm_wrappers,
     llm_response_wrapper,
-    adk_tracer_for_opik_context_management,
 )
+from .patchers.adk_otel_tracer import llm_span_helpers
 from .graph import mermaid_graph_builder
 
 LOGGER = logging.getLogger(__name__)
@@ -30,6 +30,10 @@ SpanOrTraceData = Union[span.SpanData, trace.TraceData]
 
 
 class OpikTracer:
+    """
+    Opik tracer for google-adk.
+    """
+
     def __init__(
         self,
         name: Optional[str] = None,
@@ -38,6 +42,16 @@ class OpikTracer:
         project_name: Optional[str] = None,
         distributed_headers: Optional[DistributedTraceHeadersDict] = None,
     ):
+        """
+        Initialize OpikTracer.
+
+        Arguments:
+            name: The default name for root span or trace created by the tracer.
+            tags: The default tags for all the traces and spans created by the tracer.
+            metadata: The default metadata for all the traces and spans created by the tracer.
+            project_name: The name of the project for tracing.
+            distributed_headers: The distributed trace headers.
+        """
         self.name = name
         self.tags = tags
         self.metadata = metadata or {}
@@ -160,9 +174,12 @@ class OpikTracer:
             # So we create a span manually here. This flow is handled inside ADKTracerWrapper.
             _, span_data = span_creation_handler.create_span_respecting_context(
                 start_span_arguments=arguments_helpers.StartSpanParameters(
-                    name=adk_tracer_for_opik_context_management.NAME_OF_LLM_SPAN_JUST_STARTED_FROM_OPIK_TRACER,
+                    name=model,
                     project_name=self.project_name,
-                    metadata=self.metadata,
+                    metadata={
+                        **self.metadata,
+                        llm_span_helpers.SPAN_STATUS: llm_span_helpers.LLMSpanStatus.STARTED,
+                    },
                     type="llm",
                     model=model,
                     provider=provider,
@@ -227,6 +244,11 @@ class OpikTracer:
                 usage=usage,
                 project_name=self.project_name,
             )
+            if current_span.metadata is not None:
+                current_span.metadata[llm_span_helpers.SPAN_STATUS] = (
+                    llm_span_helpers.LLMSpanStatus.READY_FOR_FINALIZATION
+                )
+
             context_storage.pop_span_data(ensure_id=current_span.id)
             current_span.init_end_time()
             # We close this span manually because otherwise ADK will close it too late,
@@ -266,6 +288,7 @@ class OpikTracer:
                 LOGGER.warning(
                     f"No current span found in context for tool: {tool.name}"
                 )
+                _log_tool_context_warning(context=tool_context)
 
         except Exception as e:
             LOGGER.error(f"Failed during before_tool_callback(): {e}", exc_info=True)
@@ -299,6 +322,7 @@ class OpikTracer:
                 LOGGER.warning(
                     f"No current span found in context for tool output update: {tool.name}"
                 )
+                _log_tool_context_warning(context=tool_context)
         except Exception as e:
             LOGGER.error(f"Failed during after_tool_callback(): {e}", exc_info=True)
 
@@ -333,3 +357,12 @@ def _try_add_agent_graph_to_metadata(
         }
     except Exception:
         LOGGER.error("Failed to build mermaid graph for agent.", exc_info=True)
+
+
+def _log_tool_context_warning(context: tool_context.ToolContext) -> None:
+    if context is not None:
+        warning = f"Function call id: {context.function_call_id}, agent name: {context.agent_name}"
+        if context.actions is not None:
+            warning += f", is escalate: {context.actions.escalate}, transfer to: {context.actions.transfer_to_agent}"
+
+        LOGGER.warning(warning)
